#include <string.h>
#include "sha256.h"
#include "pbkdf2.h"

/* Ref. RFC 2898 PKCS #5: Password-Based Cryptography Specification v.2 */

#define HS 32 /* SHA-256 output size, bytes */

static void F1(uint8_t* U, uint8_t* P, int Pn, uint8_t* S, int Sn, int i)
{
	int In = Sn + 4;
	char I[In];

	memcpy(I, S, Sn);

	uint8_t* q = (uint8_t*)(I + Sn);

	q[0] = (i >> 24) & 0xFF;
	q[1] = (i >> 16) & 0xFF;
	q[2] = (i >>  8) & 0xFF;
	q[3] = (i      ) & 0xFF;

	hmac_sha256(U, P, Pn, I, In);
}

static void Fc(uint8_t* U, uint8_t* P, int Pn)
{
	char* I = (char*) U;
	hmac_sha256(U, P, Pn, I, HS);
}

static void xorbuf(uint8_t* T, uint8_t* U, int n)
{
	for(int i = 0; i < n; i++)
		T[i] ^= U[i];
}

static void F(uint8_t* T, uint8_t* P, int Pn, uint8_t* S, int Sn, int c, int i)
{
	uint8_t U[HS];

	F1(U, P, Pn, S, Sn, i);
	memcpy(T, U, HS);

	for(int j = 2; j <= c; j++) {
		Fc(U, P, Pn);
		xorbuf(T, U, HS);
	}
}

void pbkdf2_sha256(void* psk, int len,
                 void* pass, int passlen,
                 void* salt, int saltlen, int iters)
{
	uint8_t* P = pass; int Pn = passlen;
	uint8_t* S = salt; int Sn = saltlen;

	int c = iters;
	int i;

	for(i = 1; HS*i <= len; i++) {
		uint8_t* T = psk + HS*(i-1);
		F(T, P, Pn, S, Sn, c, i);
	} if(HS*(i-1) < len) {
		uint8_t T[HS];
		F(T, P, Pn, S, Sn, c, i);
		memcpy(psk + HS*(i-1), T, len - HS*(i-1));
	}
}
